#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <limits.h>

#include "lzw.h"

#ifndef FALSE
#define FALSE	0
#endif

#ifndef TRUE
#define TRUE	(!FALSE)
#endif

#define MIN(a, b)	((a) > (b) ? (b) : (a))
#define NELEMS(x)	(sizeof(x) / sizeof(x[0]))

#define LZWMINBITS	9		/* initial number of bits per code */
#define LZWMAXBITS	12		/* unused */
#define LZWNBITS	LZWMAXBITS	/* unused */

/* control codes */
#define LZW_RESET	(LZWMINCODE+0)	/* dictionary reset */
#define LZW_EOD		(LZWMINCODE+1)	/* end-of-data */

#define LZWHASHKEY(prev, chr)	(chr | ((unsigned int)prev << 8))

#ifndef NDEBUG
#define DEBUG(fmt, arg...) do { fprintf(stderr, "debug: " fmt "\n", ##arg); } while (0)
#else
#define DEBUG(fmt, arg...) do { } while (0)
#endif

/*
 * Note that all the structures fields are considered private.
 */

/*--------------------------------------------------*/

static unsigned int djb_hash(void *buf, int len);

int
lzw_decoder_init(struct lzw_decoder *decoder)
{
	int i;

	assert(decoder != NULL);

	decoder->bitlen = LZWMINBITS;
	decoder->left = LZWMINBITS;
	decoder->code = 0;
	decoder->nextcode = LZWMINCODE + (LZWNCTLCODES-1);
	decoder->prevcode = -1;

	for (i = 0; i < NELEMS(decoder->dic); i++) {
		decoder->dic[i].chr = 0;
		decoder->dic[i].prefix = -1;
	}

	return 0;
}

int
lzw_encoder_init(struct lzw_encoder *encoder)
{
	int i;

	assert(encoder != NULL);

	encoder->bitlen = LZWMINBITS;
	encoder->left = 0;
	encoder->c = 0;
	encoder->nextcode = LZWMINCODE + (LZWNCTLCODES-1);
	encoder->ctail = 0;
	encoder->chead = 0;
	encoder->prevcode = -1;

	for (i = 0; i < LZWHASHSZ; i++)
		encoder->hash[i] = -1;

	for (i = 0; i < LZWTBLSIZE; i++) {
		encoder->dic[i].chr = 0;
		encoder->dic[i].link = -1;
		encoder->dic[i].prefix = -1;
	}

	return 0;
}

static unsigned int
djb_hash(void *buf, int len)
{
	unsigned char *s;
	unsigned int h;
	int i;

	h = 5381;
	s = buf;

	for (i = 0; i < len; i++)
		h = (h << 5) + h + s[i];
	
	return h;
}

static int
addcode(struct lzwentry *dic, short int *hash, int code, int prefix, unsigned char chr)
{
	unsigned int key;
	int idx, pos;

	assert(code <= LZWMAXCODE && (code-LZWTBLOFFS) < LZWTBLSIZE);

	if (prefix == -1)
		return 1;

	assert(code >= LZWTBLOFFS);
	pos = code - LZWTBLOFFS;
	assert(dic[pos].prefix == -1 && dic[pos].link == -1);
	dic[pos].prefix = prefix;
	dic[pos].chr = chr;

	key = LZWHASHKEY(prefix, chr);
	idx = djb_hash(&key, sizeof(key)) % LZWHASHSZ;
	dic[pos].link = hash[idx];
	hash[idx] = pos;

	return 1;
}

static int
findcode(struct lzwentry *dic, short int *hash, int prefix, unsigned char chr)
{
	unsigned int key;
	int i, idx;

	assert(dic != NULL);

	if (prefix == -1)
		return chr;

	key = LZWHASHKEY(prefix, chr);
	idx = djb_hash(&key, sizeof(key)) % LZWHASHSZ;

	for (i = hash[idx]; i >= 0; i = dic[i].link)
		if (dic[i].prefix == prefix &&
		    dic[i].chr == chr)
			return i+LZWTBLOFFS;

	return -1;
}

/* 
 * This helper emits the code. Code bits will be copied to the cbuf.
 */
static void
emitcode(struct lzw_encoder *enc, int code)
{
	unsigned int mask;
	int n;
	unsigned char c;

	n = enc->bitlen;

	/* Concatenate with the bits of the previous code, if any left. */
	if (enc->left > 0) {
		n -= enc->left;
		c = enc->c | ((unsigned int)code >> n);
		assert(enc->chead < sizeof(enc->cbuf));
		enc->cbuf[enc->chead++] = c;
		if (enc->chead >= sizeof(enc->cbuf))
			enc->chead = 0;
		enc->left = 0;
	}

	/* Add bits of the code to cbuf, save bits of the partially filled byte to output them later. */
	while (n > 0) {
		if (n >= CHAR_BIT) {
			n -= CHAR_BIT;
			c = ((unsigned int)code >> n) & 0xff;
			assert(enc->chead < sizeof(enc->cbuf));
			enc->cbuf[enc->chead++] = c;
			if (enc->chead >= sizeof(enc->cbuf))
				enc->chead = 0;
		} else {
			mask = (1 << n) - 1;
			enc->left = CHAR_BIT - n;
			enc->c = ((unsigned int)code & mask) << enc->left;
			n = 0;
		}
	}
}

int
lzw_encode(struct lzw_encoder *enc, const unsigned char **in, size_t *insz, unsigned char **out, size_t *outsz)
{
	const unsigned char *src;
	unsigned char *dst, chr;
	int code, prevcode, rc;

	assert(enc != NULL);
	assert(out != NULL && *out != NULL);
	assert(in != NULL && *in != NULL);
	assert(outsz != NULL && insz != NULL);

	rc = LZW_NOSPACE;
	src = *in;
	dst = *out;

	/* Copy any bytes to output left from the previous code bits. */
	while (enc->ctail != enc->chead) {
		if (*outsz == 0)
			break;
		*dst++ = enc->cbuf[enc->ctail++];
		if (enc->ctail >= sizeof(enc->cbuf))
			enc->ctail = 0;
		(*outsz)--;
	}

	prevcode = enc->prevcode;

	while (*insz > 0) {

		if (*outsz <= 0)
			goto err;

		/* Add next character to the string we match to known codes. */
		chr = *src;

		/* 
		 * If there is no match for the string, emit the code for the
		 * previous match and add the code for the unmatched string to
		 * the dictionary.
		 */
		if ((code = findcode(enc->dic, enc->hash, prevcode, chr)) < 0) {
			
			assert(prevcode >= 0 && prevcode != code);
			
			if (enc->nextcode <= LZWMAXCODE)
				DEBUG("encoder outputs code >> %d as part of code %d, bitlen %d", prevcode, enc->nextcode, enc->bitlen);
			else
				DEBUG("encoder outputs code >> %d", prevcode);
			/*
			 * Emit the code for the longest-known sequence. The
			 * code bits will be placed in the circle buffer.
			 */
			emitcode(enc, prevcode);
			/* 
			 * Check if we need to increase the code bit length.
			 * Output partially filled byte if we increase the bit
			 * length.
			 */
			if (enc->nextcode <= LZWMAXCODE && enc->nextcode == (1 << enc->bitlen)) {
				if (enc->left > 0) {
					enc->cbuf[enc->chead++] = enc->c;
					if (enc->chead >= sizeof(enc->cbuf))
						enc->chead = 0;
				}
				enc->bitlen++;
				enc->left = 0;
				enc->c = 0;
				DEBUG("encoder bitlen increased to %d, nextcode %d will be encoded in %d bits",
				      enc->bitlen, enc->nextcode, enc->bitlen);
			}
			/*
			 * Add the next sequence to the dictionary if it has spare codes left.
			 */
			if (enc->nextcode <= LZWMAXCODE) {
				DEBUG("encoder adds new code %d for prevcode %d char %c", enc->nextcode, prevcode, chr);
				assert(enc->nextcode != prevcode);
				addcode(enc->dic, enc->hash, enc->nextcode, prevcode, chr);
				enc->nextcode++;
			}
			/* Write encoded output to the buffer. */
			while (enc->ctail != enc->chead) {
				if (*outsz == 0)
					break;
				*dst++ = enc->cbuf[enc->ctail++];
				if (enc->ctail >= sizeof(enc->cbuf))
					enc->ctail = 0;
				(*outsz)--;
			}
			prevcode = chr;
		} else {
			prevcode = code;
		}

		/* Proceed to the next input byte. */
		(*insz)--;
		src++;
	}

	rc = LZW_OK;
err:
	enc->prevcode = prevcode;
	*out = dst;
	*in = src;
	return rc;
}

int
lzw_end_of_data(struct lzw_encoder *enc, unsigned char **out, size_t *outsz)
{
	unsigned char *dst;
	unsigned int count;
	int rc;

	assert(enc != NULL && out != NULL && outsz != NULL);

	/*
	 * Check the case when no finalization needed. It is possible
	 * that the user already did the finalization.
	 */
	if (enc->prevcode == -1 && enc->ctail == enc->chead)
		return LZW_OK;

	rc = LZW_NOSPACE;
	dst = *out;

	/* Copy any bytes to output left from the previous code bits. */
	while (enc->ctail != enc->chead) {
		if (*outsz == 0)
			goto err;
		*dst++ = enc->cbuf[enc->ctail++];
		if (enc->ctail >= sizeof(enc->cbuf))
			enc->ctail = 0;
		(*outsz)--;
	}

	/*
	 * At this point, circle buffer was emptied completely. Next code
	 * bits will be produced at the beginning of the circle buffer to
	 * simplify handling the finalization.
	 */
	enc->ctail = enc->chead = 0;
	
	if (enc->prevcode >= 0) {
		int bitlen;
		int left;
		unsigned char c;

		/* 
		 * Save encoder state before code emission. The state of the
		 * encoder will be restored back if the output buffer provided
		 * by the user has not enough space to copy the code and LZW_EOD
		 * simultaneously.
		 */
		bitlen = enc->bitlen;
		left = enc->left;
		c = enc->c;

		assert(enc->prevcode >= 0);
		emitcode(enc, enc->prevcode);

		/*
		 * It is possible that emission of the code will require
		 * us to increment the bit length before we emit LZW_EOD.
		 * Carefully check and handle this case.
		 */
		if (enc->nextcode <= LZWMAXCODE && enc->nextcode == (1 << enc->bitlen)) {
			if (enc->left > 0)
				enc->cbuf[enc->chead++] = enc->c;
			enc->bitlen++;
			enc->left = 0;
			enc->c = 0;
			DEBUG("encoder bitlen increment in finalize to %d", enc->bitlen);
		}
		/* 
		 * Emission of LZW_EOD will reset the decoder's previous
		 * sequence collected in the buffer. It prevents it to be used
		 * to construct a new sequence to add to the dictionary.
		 */
		emitcode(enc, LZW_EOD);

		/* Put partially filled byte of LZW_EOD code to output. */
		if (enc->left > 0) {
			enc->cbuf[enc->chead++] = enc->c;
			enc->left = 0;
		}
		
		assert(enc->chead > enc->ctail);
		if ((count = enc->chead - enc->ctail) > *outsz) {
			/* 
			 * Restore encoder state back. This allows a user to
			 * increase the output buffer size and repeat output
			 * finalization again.
			 */
			enc->ctail = enc->chead = 0;
			enc->c = c;
			enc->left = left;
			enc->bitlen = bitlen;
			goto err;
		}
		
		memcpy(dst, enc->cbuf, count);
		*outsz -= count;
		dst += count;

		enc->ctail = enc->chead = 0;
	}

	enc->prevcode = -1;
	rc = LZW_OK;
err:	
	*out = dst;
	return rc;
}

int
lzw_reset(struct lzw_encoder *enc, unsigned char **out, size_t *outsz)
{
	unsigned char *dst;
	int count, rc;

	assert(enc != NULL);
	assert(out != NULL && outsz != NULL);

	/* Check if the user finalized data transmission. */
	if (enc->prevcode != -1)
		return LZW_FAIL;

	assert(enc->ctail == enc->chead);

	rc = LZW_NOSPACE;
	dst = *out;

	emitcode(enc, LZW_RESET);

	/* Put partially filled byte of LZW_RESET code to circle buffer. */
	if (enc->left > 0) {
	 	enc->cbuf[enc->chead++] = enc->c;
		enc->left = 0;
		enc->c = 0;
	}

	/* Try to output the circle buffer all at once. */
	if ((count = enc->chead - enc->ctail) <= *outsz) {
		int i;

		memcpy(dst, enc->cbuf, count);
		*outsz -= count;
		dst += count;

		enc->bitlen = LZWMINBITS;
		enc->left = 0;
		enc->c = 0;
		enc->nextcode = LZWMINCODE + (LZWNCTLCODES-1);
		enc->ctail = enc->chead = 0;
		enc->prevcode = -1;

		for (i = 0; i < LZWTBLSIZE; i++) {
			enc->dic[i].chr = 0;
			enc->dic[i].link = -1;
			enc->dic[i].prefix = -1;
		}

		for (i = 0; i < LZWHASHSZ; i++)
			enc->hash[i] = -1;

		rc = LZW_OK;
	}

	*out = dst;
	return rc;
}

int
lzw_complete(struct lzw_encoder *enc)
{
	assert(enc != NULL);

	return enc->nextcode > LZWMAXCODE;
}

/* This helper resets the decoder previous sequence. */
static void
eod(struct lzw_decoder *decoder)
{
	DEBUG("decoder EOD");
	decoder->prevcode = -1;
	decoder->left = decoder->bitlen;
	decoder->code = 0;
}

/* This helper clears the dictionary and restores the decoder to the initial state. */
static void
reset(struct lzw_decoder *decoder)
{
	struct lzwseq *dic;
	int i;

	dic = decoder->dic;

	for (i = 0; i < NELEMS(decoder->dic); i++) {
		dic[i].chr = 0;
		dic[i].prefix = -1;
	}

	decoder->bitlen = LZWMINBITS;
	decoder->left = decoder->bitlen;
	decoder->code = 0;
	decoder->prevcode = -1;
	decoder->nextcode = LZWMINCODE + (LZWNCTLCODES-1);
}

/* This helper adds the code to the dictionary. */
static int
addseq(struct lzwseq *dic, int code, int prefix, char chr)
{
	assert(dic != NULL && code >= 0 && code <= LZWMAXCODE);
	assert(code >= LZWTBLOFFS);
	assert(dic[code-LZWTBLOFFS].prefix == -1);

	dic[code-LZWTBLOFFS].prefix = prefix;
	dic[code-LZWTBLOFFS].chr = chr;
	return 0;
}

/* This helper reconstructs the sequence from its code. */
static const unsigned char *
reconstr(struct lzwseq *dic, int code, unsigned char buf[LZWMAXBUF], int *len)
{
	unsigned char *cp;
	int idx;

	assert(dic != NULL && len != NULL && buf != NULL);
	assert(code >= LZWTBLOFFS && code <= LZWMAXCODE);

	cp = buf + LZWMAXBUF-1;
	idx = code - LZWTBLOFFS;

	while (idx >= 0) {
		assert(cp >= buf);
		*cp-- = dic[idx].chr;
		if (dic[idx].prefix < LZWMINCODE) {
			assert(cp >= buf);
			*cp = (unsigned char)dic[idx].prefix;
			break;
		}
		idx = dic[idx].prefix - LZWTBLOFFS;
	}

	*len = buf + LZWMAXBUF - cp;
	assert(*len <= LZWMAXBUF && *len > 0);

	return cp;
}
	
static int
putcode(struct lzw_decoder *dec, int code, unsigned char **out, size_t *outsz)
{
	unsigned char buf[LZWMAXBUF];
	unsigned char *dst;
	struct lzwseq *dic;
	const unsigned char *cp;
	size_t size;
	int code_exists, len;
	int rc;
	
	assert(code >= 0 && code <= LZWMAXCODE);

	rc = LZW_FAIL;
	dst = *out;
	size = *outsz;

	len = 0;
	code_exists = FALSE;
	dic = dec->dic;

	if (code >= LZWMINCODE) {
		assert(code >= LZWTBLOFFS);
		if (dic[code-LZWTBLOFFS].prefix >= 0)
			code_exists = TRUE;
	} else {
		code_exists = TRUE;
		buf[0] = (unsigned char)code;
		len = 1;
	}

	if (code_exists) {

		if (len == 0) {
			cp = reconstr(dic, code, buf, &len);
			assert(cp != NULL && len > 0);
		} else
			cp = buf;

		/* When no output buffer space left, return LZW_NOSPACE. */
		if (len > size) {
			rc = LZW_NOSPACE;
			goto err;
		}

		DEBUG("decoder outputs code >> %d, bitlen %d", code, dec->bitlen);

		if (dec->prevcode >= 0 && dec->nextcode <= LZWMAXCODE) {
			addseq(dic, dec->nextcode, dec->prevcode, cp[0]);
			DEBUG("decoder adds code %d for sequence %d+%c", dec->nextcode, dec->prevcode, cp[0]);
			dec->nextcode++;
		}

		/* Copy decoded sequence to the output buffer. */
		memcpy(dst, cp, len);
		size -= len;
		dst += len;

		/* 
		 * Save the code, later we will synthesize a next code sequence out of it.
		 */
		dec->prevcode = code;
	} else {
		if (dec->prevcode >= 0 && code == dec->nextcode) {
		
			/*
			 * Special treatment required if we get an unknown code with
			 * the index equal to the next unused dictionary index.
			 */
			if (dec->prevcode >= LZWMINCODE) {
				len = 0;
				cp = reconstr(dic, dec->prevcode, buf, &len);
				assert(cp != NULL && len > 0);
			} else {
				buf[0] = (unsigned char)dec->prevcode;
				cp = buf;
				len = 1;
			}

			if (len+1 > size) {
				rc = LZW_NOSPACE;
				goto err;
			}

			addseq(dic, dec->nextcode, dec->prevcode, cp[0]);
			dec->nextcode++;

			memcpy(dst, cp, len);
			dst += len;
			size -= len+1;
			*dst++ = cp[0];
			
			dec->prevcode = code;
		} else {
			if (code >= LZWMINCODE+LZWNCTLCODES) {
				rc = LZW_BADCODE;
				goto err;
			}
		}
	}

	rc = LZW_OK;
err:
	*outsz = size;
	*out = dst;

	return rc;
}

int
lzw_decode(struct lzw_decoder *dec, const unsigned char **in, size_t *insz, unsigned char **out, size_t *outsz)
{
	const unsigned char *buf;
	unsigned int left, mask;
	int code, rc;
	size_t size;
	unsigned char c;

	assert(dec != NULL);
	assert(in != NULL && insz != NULL);
	assert(out != NULL && outsz != NULL);

	buf = *in;
	size = *insz;
	rc = LZW_NOSPACE;

	while (size > 0) {

		c = *buf;

		DEBUG("decoder processing c=%02x, from which I'll consume %d bits", c, dec->left > CHAR_BIT ? CHAR_BIT: dec->left);

		/* Consume next input byte, adding it's bits to the code. */
		if (dec->left > CHAR_BIT) {
			left = dec->left - CHAR_BIT;
			code = dec->code | (c << left);
			if (left == 0) {
				switch (code) {
				case LZW_RESET:
					reset(dec);
					break;
				case LZW_EOD:
					eod(dec);
					break;
				default:
					/*
					 * If putcode() was not ok, keep the decoder state intact.
					 */
					if ((rc = putcode(dec, code, out, outsz)) != LZW_OK)
						goto err;
					break;
				}
				dec->left = 0;
				dec->code = 0;
			} else {
				dec->left = left;
				dec->code = code;
			}
		} else {
			assert(dec->left <= CHAR_BIT);
			code = dec->code | (c >> (CHAR_BIT - dec->left));
			switch (code) {
			case LZW_RESET:
				reset(dec);
				break;
			case LZW_EOD:
				eod(dec);
				break;
			default:
				/*
				 * Decoder state is not changed if putcode()
				 * failed.
				 */
				if ((rc = putcode(dec, code, out, outsz)) != LZW_OK)
					goto err;
				mask = (1 << (CHAR_BIT - dec->left))-1;
				dec->left = dec->bitlen - (CHAR_BIT - dec->left);
				dec->code = (c & mask) << dec->left;
				break;
			}
		}
		
		if (dec->nextcode <= LZWMAXCODE && dec->nextcode == (1 << dec->bitlen)) {
			dec->code = 0;
			dec->bitlen++;
			dec->left = dec->bitlen;
			DEBUG("decoder increased bitlen to %d, nextcode %d will read %d bits", dec->bitlen, dec->nextcode, dec->bitlen);
		}

		DEBUG("---> next byte");

		/* Skip to the next input byte. */
		size--;
		buf++;
	}

	rc = LZW_OK;
err:
	*insz = size;
	*in = buf;
	return rc;
}

